/*
 * sampleCmd.c --
 * Copyright (c) 1997 Clif Flynt
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL Clif Flynt BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Clif Flynt SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND Clif Flynt HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/
/* Include the usual suspects.						*/

#include "stdlib.h"
#include "ctype.h"
#include "sampleInt.h"

#define SAMPLECMD_REVISION "$Revision: 1.12 $"

/*
* Define the sub commands                          
*
* These strings define the subcommands that the sample command 
* supports.
*
* To add a new subcommand, add the new subcommand string, 
* #define and entry in cmdDefinition
*
* Note: Order is important.  
*
* These are the subcommands that will be recognized
*
*/

static char *subcommands[] = {
	"create", "set", "get", "debug", "destroy", "liststr",
	"arraystr", "getstr", "setvar", "bogus", NULL};

/*
* These #defines define the positions of the subcommands.
* You can use enums if you are certain your compiler will provide the
* same numbers as this.
*/

#define M_create 0
#define M_set    1
#define M_get    2
#define M_debug  3
#define M_destroy 4
#define M_liststr 5
#define M_arraystr 6
#define M_getstr 7
#define M_setvar    8
#define M_bogus 9

/*
* The cmdDefinition structure describes the minimum and maximum number
*  of expected arguments for the subcommand (including cmd and subcommand
*  names), and a usage message to return if the argument
*  count is outside the expected range.
*/

typedef struct cmd_Def {
   char *usage;
   int  minArgCnt;
   int  maxArgCnt;
   } cmdDefinition;

static cmdDefinition definitions[] = {
    {"create key raw_message", 4 , 4},
    {"set arrayName index Value", 5, 5},
    {"get key ", 3, 3},
    {"debug level", 3, 3},
    {"destroy key", 3,3},
    {"liststr listName ", 4, 4},
    {"arraystr arrayName", 3,3},
    {"getstr key", 3,3},
    {"setvar var key", 4 ,4},
    {"bogus bogusval", 3,3}
};

/* If sampleDebugPrint != 0, then debugprt will print debugging info */
/*   This value is set with the default subcommand function debug   */

int sampleDebugPrint = 0;

/* ----- C++ Comment
// These are the segments that need to be declared in the "C" namespace
// These function names and data will be accessed by programs in the
//   Tcl "C" language library.
*/

/*
extern "C" {
*/

/*
 *----------------------------------------------------------------------
 *
 * Sample_Cmd --
 *
 *      Sample_Cmd is invoked to process the "sample" Tcl command.
 *      It will parse a subcommand, and perform the requested action.
 *   
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 * 
 *----------------------------------------------------------------------
 */

int Sample_Cmd(ClientData sample, 
             Tcl_Interp *interp, 
	     int objc, 
	     Tcl_Obj * CONST *objv) {

    /* ClientData sample;           /* Not used. */
    /* Tcl_Interp *interp;         /* Current interpreter. */
    /* int objc;                   /* Number of arguments. */
    /* Tcl_Obj *CONST objv[];      /* Argument objects. */

  int cmdnum;
  int result;
  Tcl_Obj *returnValue;
  CmdReturn *returnStruct;
  programInfo info;
  
  /*
   * Initialize the return value
   */

  returnValue = NULL;
  returnStruct = NULL;
  
  /*
   *  Check that we have at least a subcommand, 
   *   else return an Error and the usage message 
   */

  if (objc < 2) {
      Tcl_WrongNumArgs(interp, 1, objv, 
              "subcommand ?sampleID? ?fieldName? ?fieldValue?");
      return TCL_ERROR;
  }

  /*
   * Find this sample subcommand in the list of subcommands.  
   * Tcl_GetIndexFromObj returns the offset of the recognized string,
   * which is used to index into the command definitions table.
   */

  result = Tcl_GetIndexFromObj(interp, objv[1], subcommands, 
  	                       "subcommand", TCL_EXACT, &cmdnum);
   
  /* 
   * If the result is not TCL_OK, then the error message is already
   *    in the Tcl Interpreter, this code can immediately return.
   */

  if (result != TCL_OK) {
      return TCL_ERROR;
  }

  /*
   *  Check that the argument count matches what's expected for this
   * Subcommand.
   */

  if ((objc < definitions[cmdnum].minArgCnt) || 
      (objc > definitions[cmdnum].maxArgCnt) ) {
      Tcl_WrongNumArgs(interp, 1, objv, definitions[cmdnum].usage);
      return TCL_ERROR;
  }
    
			
  result = TCL_OK;

  /* 
   * The subcommand is recognized, and has a valid number of arguments
   * Process the command.
   */

  switch (cmdnum) {
      case M_debug: {
    		char *tmp;
		tmp = Tcl_GetStringFromObj(objv[2], NULL);
		if (TCL_OK != Tcl_GetInt(interp, tmp, &sampleDebugPrint)) {
		    return (TCL_ERROR);
		}
		break;
      }
      case M_destroy: {
		returnStruct = 
		    Sample_DestroyCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_create:  {
		returnStruct = 
		    Sample_CreateCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_get:	    {
	    	returnStruct = 
		    Sample_GetCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_set:	    {
	    	returnStruct = 
		    Sample_SetCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_setvar:	    {
	    	returnStruct = 
		    Sample_SetVarCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_liststr:	    {
	    	returnStruct = 
		    Sample_ListstrCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_arraystr:	    {
	    	returnStruct = 
		    Sample_ArraystrCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      case M_getstr:	    {
	    	returnStruct = 
		    Sample_GetstrCmd((ClientData) &info, interp, objc, objv);
		break;
      }
      default:		   {
      		char error[80];
		sprintf(error, "Bad sub-command %s.  Has no entry in switch",
		    Tcl_GetStringFromObj(objv[1], NULL));
		returnValue = Tcl_NewStringObj(error, -1);
		result = TCL_ERROR;
      }
  }
  
  /* 
   * Extract an object to return from returnStruc.
   * returnStruct will be NULL if the processing is done in this 
   *  function and no other function is called.
   */

  if (returnStruct != NULL) {
      returnValue = returnStruct->object;
      result = returnStruct->status;
      free (returnStruct);
  }
  
  /* 
   * Set the return value and return the status
   */

  if (returnValue != NULL) {
      Tcl_SetObjResult(interp, returnValue);
  }

  return result;
}
/* ----- C++ Comment
// This is the close of the extern "C" section
*/

/*
 }
*/



//
// A set of simple examples of command usage.
//  These do not include command line checking, error checking
//  or other useful (required!) features.
//

// Example of setting an array variable 
//  sample::Set      name     index    value
//  objv[0]          objv[1]  objv[2]  objv[3]

int Simple_SetVar2Cmd (
          ClientData info, Tcl_Interp *interp, int objc,
          Tcl_Obj * CONST * objv) {
  Tcl_Obj *returnObjPtr;

  returnObjPtr = 
     Tcl_ObjSetVar2(interp, objv[1], objv[2], objv[3], 0);
    
  Tcl_SetObjResult(interp, returnObjPtr);
  return TCL_OK;
}

// Simple return of the value of an array var
//  sample::Get name     index
//  objv[0]     objv[1]  objv[2]

int Simple_GetVar2Cmd (
          ClientData info, Tcl_Interp *interp, int objc,
          Tcl_Obj * CONST * objv) {
  Tcl_Obj *returnObjPtr;

  returnObjPtr = 
     Tcl_ObjGetVar2(interp, objv[1], objv[2], 0);
    
  Tcl_SetObjResult(interp, returnObjPtr);
  return TCL_OK;
}


// Return a Tcl list of numbers 1, 2, 3
// sample::MakeList
// objv[0]

int Simple_ListMake (
          ClientData info, Tcl_Interp *interp, int objc,
          Tcl_Obj * CONST * objv) {


  Tcl_Obj *returnObjPtr;
  Tcl_Obj *listPtrPtr[3];
  int i;
  
  for(i=0; i < 3; i++) {
     listPtrPtr[i] = Tcl_NewIntObj(i);
  }
  
  returnObjPtr = Tcl_NewListObj(3, listPtrPtr);
  
  Tcl_SetObjResult(interp, returnObjPtr);
  return TCL_OK;
}

// Demonstrate the Tcl_IsShared function.
//
// sample::Share [list a b c] -- returns 0
//   The object is created for this command, but not used by any
//   Tcl variable.
// sample::Share [set x 2] -- returns 1
//   The object is used by a tcl variable (x)

int Simple_Share  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {
  Tcl_Obj *returnObjPtr;
  returnObjPtr = Tcl_NewIntObj(Tcl_IsShared(objv[1]));
  Tcl_SetObjResult(interp, returnObjPtr);
  return TCL_OK;
}   
 
// Return a new list with a value appended to the original 
//  list
//
// sample::Append list     val
// objv[0]        objv[1]  objv[2]

Simple_AppendList (
          ClientData info, Tcl_Interp *interp, int objc,
          Tcl_Obj * CONST * objv) {

  Tcl_Obj *returnObjPtr, *newObj;
  int len, rslt;
  
  Tcl_ListObjLength(interp, objv[1], &len);
  
  newObj = Tcl_DuplicateObj(objv[1]);

  rslt = Tcl_ListObjAppendElement(interp, newObj, objv[2]);
  
  Tcl_SetObjResult(interp, newObj);

  Tcl_ListObjLength(interp, objv[1], &len);

  return rslt;
}

// Append a value to a list - modify the list
//  equivalent to the lappend command
//
// sample::Mod  list     newval
// objv[0]      objv[1]  objv[2]

Simple_ModList (
          ClientData info, Tcl_Interp *interp, int objc,
          Tcl_Obj * CONST * objv) {

  Tcl_Obj *returnObjPtr, *newObj;
  int len, rslt;
  
  Tcl_ListObjLength(interp, objv[1], &len);
  
  Tcl_DecrRefCount(objv[1]);
  newObj = objv[1];

  rslt = Tcl_ListObjAppendElement(interp, newObj, objv[2]);

  Tcl_IncrRefCount(objv[1]);
  
  Tcl_SetObjResult(interp, newObj);

  Tcl_ListObjLength(interp, objv[1], &len);

  return rslt;
}

// Display the length of a list
// sample::ListLen list 
// objv[0] objv[1]

int Simple_ListLen  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {
  Tcl_Obj *returnObjPtr;
  int len;
  Tcl_ListObjLength(interp, objv[1], &len);
  returnObjPtr = Tcl_NewIntObj(len);
  Tcl_SetObjResult(interp, returnObjPtr);
  return TCL_OK;
}   

// Return list element at an index 
//  lindex equivalent
// sample::ListIndex list    index
// objv[0]           objv[1] objv[2]

int Simple_ListIndex  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {
  Tcl_Obj *returnObjPtr;
  int index;

  Tcl_GetIntFromObj(interp, objv[2], &index);
  Tcl_ListObjIndex(interp, objv[1], index, &returnObjPtr);

  Tcl_SetObjResult(interp, returnObjPtr);

  return TCL_OK;
}   

// Create a new variable - naemd newVariable in current scope.
// sample::MakeVar
// No return value.
// objv[0]

int Simple_MakeVar  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {

  Tcl_SetVar(interp, "newVariable", "New Data In Var", 0);

  return TCL_OK;
}   

// Demonstrates accessing a Tcl variable from C with Tcl_GetVar
// sample::PrintVar name
// objv[0]          objv[1]

int Simple_PrintVar  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {
  char *val, *name;

  name = Tcl_GetStringFromObj(objv[1], NULL);
  val = Tcl_GetVar(interp, name, 0);
  printf("VAL: %s\n", val);
  return TCL_OK;
}   

// Demonstrate accessing a Tcl variable from C with Tcl_ObjGetVar2
// sample::PrintVar2  name      index
// objv[0]            objv[1]   objv[2]

int Simple_PrintVar2  (ClientData info, Tcl_Interp *interp,
              int objc, Tcl_Obj * CONST *objv) {
  Tcl_Obj *objPtr, *name2;
  char *val;

  if (objc == 2) {
      name2 = NULL;
  } else {
      name2 = objv[2];
  }
  objPtr = Tcl_ObjGetVar2(interp, objv[1], name2, 0);
  val = Tcl_GetStringFromObj(objPtr, NULL);
  printf("VAL: %s\n", val);
  
  return TCL_OK;
}   

