/*
 * sampleCmd.c --
 * Copyright (c) 1997 Clif Flynt
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL Clif Flynt BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Clif Flynt SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND Clif Flynt HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/
/*  Include the usual suspects.                    */

#include "stdlib.h"
#include "ctype.h"
#include "sampleInt.h"


static Tcl_HashTable *sample_hashtablePtr;
extern int sampleDebugPrint;

#define SAMPLECMDAZ_REVISION "$Revision: 1.15 $"

/*------------------------------------------------------------------------
* void Sample_InitHashTable ()--
*    Initialize a hash table.
*    If your application does not need a hash table, this may be deleted.
*
* Arguments
*   NONE
* 
* Results
*   Initializes the hash table to accept STRING keys
*
* Side Effects:
*   None
------------------------------------------------------------------------*/

void Sample_InitHashTable () {
    sample_hashtablePtr = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(sample_hashtablePtr, TCL_STRING_KEYS);
}

/*------------------------------------------------------------------------
* CmdReturn *Sample_CreateCmd ()--
*    Demonstrates creating a hash entry.
*    Creates a hash entry for Key, with value String
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"create"
*   objv[2]:	hash Key
*   objv[3]:	String
* 
* Results
*   Creates a new hash entry.  Sets error if entry already exists.
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_CreateCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    Tcl_HashEntry *hashEntryPtr;
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    char *tmpString;
    Tcl_Obj *returnObjPtr;

    char *hashEntryContentsPtr;
    char *hashKeyPtr;
    int  isNew;
    int length;
    
    /*
     * Print that the function was called for debugging
     */

    debugprt("Sample_CreateCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /* 
     * Extract the string representation of the object argument,
     * and use that as a key into the hash table.
     *
     * If this entry already exists, complain.
     */

    hashKeyPtr = Tcl_GetStringFromObj(objv[2], NULL);
    hashEntryPtr = Tcl_CreateHashEntry(sample_hashtablePtr, hashKeyPtr, &isNew);
    
    if (!isNew) {
        char errString[80];

	sprintf(errString, "Hashed object named \"%s\" already exists.\n", 
	    hashKeyPtr);

        /*
         * Both of these strings will be added to the Tcl script
	 *  global variable errorInfo
	 */

        Tcl_AddObjErrorInfo(interp, errString, strlen(errString));
        Tcl_AddErrorInfo(interp, "error in Sample_CreateCmd");
	
	/*
	 * This SetErrorCode command will set the Tcl script
	 *  variable errorCode to "5 5 5"
	 */

	Tcl_SetErrorCode(interp, "APPLICATION", "Name in Use",(char *) NULL);
	
	/* 
	 * This defines the return string for this subcommand
	 */

        Tcl_AppendResult(interp, "Hashed object named \"",
                hashKeyPtr, "\" already exists.", (char *) NULL);

	returnStructPtr->status = TCL_ERROR;
        goto done;
    }
    
    /*
     * If we are here, then the key is unused.
     * Get the string representation from the object, 
     *  and make a copy that can be placed into the hash table.
     */

    tmpString = Tcl_GetStringFromObj(objv[3], &length);
    hashEntryContentsPtr = (char *) malloc(length+1);
    strcpy(hashEntryContentsPtr, tmpString);
    
    debugprt("setting: %s\n", hashEntryContentsPtr);
    Tcl_SetHashValue(hashEntryPtr, hashEntryContentsPtr);

done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}

/*------------------------------------------------------------------------
* CmdReturn *Sample_GetCmd ()--
*    Demonstrates retrieving a value from a hash table.
*    Returns the value of the requested item in the hash table
*
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"get"
*   objv[2]:	hash Key
* 
* Results
*   No changes to hash table.  Returns saved value, or sets error.
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_GetCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    Tcl_HashEntry *hashEntryPtr;
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    Tcl_Obj *returnObjPtr;
    
    char *hashKeyPtr;
    
    debugprt("Sample_GetCmd called with %d args\n", objc);
    
    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = NULL;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /*
     * Get the key from the argument
     * And attempt to extract that entry from the hashtable.
     * If the returned entry pointer is NULL, this key is not in the table.
     */

    hashKeyPtr = Tcl_GetStringFromObj(objv[2], NULL);

    hashEntryPtr = Tcl_FindHashEntry(sample_hashtablePtr, hashKeyPtr);

    if (hashEntryPtr == (Tcl_HashEntry *) NULL) {
        char errString[80];
	Tcl_Obj *errCodePtr;
	Tcl_Obj *objv[2];

	/* 
         * Define an error code as a list.  Set errorCode.        
	 */

        objv[0] = Tcl_NewStringObj("APPLICATION", -1);
        objv[1] = Tcl_NewStringObj("No such name", -1);

	errCodePtr = Tcl_NewListObj(2, objv);
	Tcl_SetObjErrorCode(interp, errCodePtr);
	
	/* 
	 * This string will be placed in the global variable errorInfo
	 */
	
	sprintf(errString, "Hash object \"%s\" does not exist.", hashKeyPtr);
        Tcl_AddErrorInfo(interp, errString);
	
	/* 
	 * This string will be returned as the result of the command.
	 */

        Tcl_AppendResult(interp, "can not find hashed object named \"",
                hashKeyPtr, "\"", (char *) NULL);
	returnStructPtr->status = TCL_ERROR;
        goto done;
    }
    
    /* 
     * If we got here, then the search was successful and we can extract
     *  the data value from the hash entry and return it.
     */

    returnCharPtr = (char *)Tcl_GetHashValue(hashEntryPtr);
    
    debugprt("returnString: %s\n", returnCharPtr);

done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}



/*------------------------------------------------------------------------
* CmdReturn *Sample_DestroyCmd ()--
*    Demonstrate destroying an entry in the hash table.
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"destroy"
*   objv[2]:	hash Key
* 
* Results
*   Deletes the hash table entry and frees the memory associated with
*      the hash object being stored.
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_DestroyCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    Tcl_HashEntry *hashEntryPtr;
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    Tcl_Obj *returnObjPtr;

    char *hashEntryContentsPtr;
    char *hashKeyPtr;

    debugprt("Sample_DestroyCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = NULL;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /* 
     * Extract the string representation from the argument, and use
     * it as a key into the hash table.
     */
    
    hashKeyPtr = Tcl_GetStringFromObj(objv[2], NULL);
    hashEntryPtr = Tcl_FindHashEntry(sample_hashtablePtr, hashKeyPtr);
    
    /*
     * If the hashEntryPtr returns NULL, then this key is not in
     *  the table.  Return an error.
     */

    if (hashEntryPtr == (Tcl_HashEntry *) NULL) {
        /*
	 * Tcl_AppendResult sets the return for this command.
	 * The script global variable errorInfo will also be set to
	 *   this string.
	 * The script global variable errorCode will be set to "NONE"
	 */

        Tcl_AppendResult(interp, "can not find hashed object named \"",
                hashKeyPtr, "\"", (char *) NULL);
	returnStructPtr->status = TCL_ERROR;
        goto done;
    }
    
    hashEntryContentsPtr = (char *)Tcl_GetHashValue(hashEntryPtr);
    free(hashEntryContentsPtr);

    Tcl_DeleteHashEntry(hashEntryPtr);
done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}


/*------------------------------------------------------------------------
* CmdReturn *Sample_SetCmd ()--
*   Demonstrates setting an array to a value
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"set"
*   objv[2]:	arrayName
*   objv[3]:	Index
*   objv[4]:	New Value
* 
* Results
*   Sets arrayName(Index) to the Value  
*   Also sets arrayName(alpha) to the string "NewValue".
*   Returns the string "NewValue"
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_SetCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST * objv) {
    Tcl_Obj *returnObjPtr;
    Tcl_Obj *indexObjPtr; 
    Tcl_Obj *arrayObjPtr;
    Tcl_Obj *valueObjPtr;
    CmdReturn *returnStructPtr;
    
    debugprt("Sample_SetCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */
    
    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnObjPtr    = NULL;
    
    /*
     * Use Tcl_ObjSetVar2 to set the array element "Index" to "Value"
     */
    
    arrayObjPtr = objv[2];
    Tcl_ObjSetVar2(interp, arrayObjPtr, objv[3], objv[4], 0);
    
    /*
     * Create two new objects to set a value for index "alpha" in
     *  the array.
     */

    indexObjPtr = Tcl_NewStringObj("alpha", -1);
    valueObjPtr = Tcl_NewStringObj("NewValue", -1);
    returnObjPtr = Tcl_ObjSetVar2(interp, arrayObjPtr, indexObjPtr, valueObjPtr, 0);
    
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = returnObjPtr;

    /* 
     * Delete the temporary objects by reducing their RefCount 		
     *   The object manager will free them, and associated memory when  
     *   the reference count becomes 0.           			
     */

    Tcl_DecrRefCount(indexObjPtr);

    /*	 Don't delete valueObjPtr - it's the returnObjPtr object. 
     *   The task will core dump if you clear it, and then use it.	
     *   Tcl_DecrRefCount(valueObjPtr);					
     */

    return(returnStructPtr);
    }

/*------------------------------------------------------------------------
* CmdReturn *Sample_SetVarCmd ()--
*   Demonstrates setting a named variable to a value
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"set"
*   objv[2]:	varName
*   objv[3]:	New Value
* 
* Results
*   Sets arrayName(Index) to the Value  
*   Also sets arrayName(alpha) to the string "NewValue".
*   Returns the string "NewValue"
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_SetVarCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST * objv) {
    Tcl_Obj *returnObjPtr;
    Tcl_Obj *valueObjPtr;
    CmdReturn *returnStructPtr;
    
    debugprt("Sample_SetCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */
    
    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnObjPtr    = NULL;
    
    /*
     * Use Tcl_ObjSetVar2 to set the array element "Index" to "Value"
     */
    printf("OBJ2: %s - ", Tcl_GetStringFromObj(objv[2], NULL));    
    printf("OBJ3: %s \n", Tcl_GetStringFromObj(objv[3], NULL));    


    Tcl_SetVar(interp, Tcl_GetStringFromObj(objv[2], NULL), 
        Tcl_GetStringFromObj(objv[3], NULL),  0);
	
    returnObjPtr = objv[3];
    
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = returnObjPtr;

    return(returnStructPtr);
    }



/*------------------------------------------------------------------------
* CmdReturn *Sample_ArraystrCmd (ClientData info--
*    Extracts fields from an array to fill a structure.
* Arguments:
*   objv[0]:	"sample"
*   objv[1]:	"arraystr"
*   objv[2]:	arrayname
* Results:
*   Prints out the contents of the structure after filling it.
* 
* Side Effects:
*   None
*
------------------------------------------------------------------------*/
CmdReturn *Sample_ArraystrCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    Tcl_Obj *returnObjPtr;
    Tcl_Obj *structElementPtr;
    Tcl_Obj *indexPtr;
    int i;
    
    struct demo {
        int	firstInt;
	char    *secondString;
	double  thirdFloat;
    } demoStruct;
    
    /* The names of the fields that will be used as array indices */

    char *fields[] = {"firstInt", "secondString", "thirdFloat", NULL };

    debugprt("Sample_ArraystrCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = NULL;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /*
     * Create an object that can be used with Tcl_ObjGetVar2 to
     *  find the object identified with the array and index names
     */

    indexPtr = Tcl_NewStringObj(fields[0], -1);
    
    for (i=0; i<3; i++) {

        Tcl_SetStringObj (indexPtr, fields[i], -1);
	
	/* 
	 * Get the object identified as arrayName(index) 
	 * If that value is NULL, there is no arrayName(index): complain.
	 */

        structElementPtr = Tcl_ObjGetVar2(interp, objv[2], indexPtr, 0);

	if (structElementPtr == NULL) {
            Tcl_AppendResult(interp, "Array Index \"", fields[i],
                 "\" is not defined", (char *) NULL);
            returnStructPtr->status = TCL_ERROR;
            goto done;
	}
        
	/* 
	 * A strange way to determine which structure element is
	 *  being processed, but it works in the loop.
	 * If an illegal value is in the object, the error return
	 *  will be set automaticly by the interpreter.
	 */

	switch (i) {
	case 0: {
	            int t;
		    if (TCL_OK !=
		        Tcl_GetIntFromObj(interp, structElementPtr, &t)) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct.firstInt = t;
		    }
		break;
		}
	case 1: {
	            char *t;
		    if (NULL == (t = 
		        Tcl_GetStringFromObj(structElementPtr, NULL))) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct.secondString= (char *)malloc(strlen(t)+1);
			strcpy(demoStruct.secondString, t);
		    }
		break;
		}
	case 2: {
	            double t;
		    if (TCL_OK !=
		        Tcl_GetDoubleFromObj(interp, structElementPtr, &t)) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct.thirdFloat = t;
		    }
		break;
		}
	}
    }

/* 
 * Display the contents of the structure as proof of having extracted
 * the fields.
 */

printf("demoStruct assigned values: \n %d\n %s\n %f\n", 
  demoStruct.firstInt, demoStruct.secondString, demoStruct.thirdFloat);
    
done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}

/*------------------------------------------------------------------------
* CmdReturn *Sample_ListstrCmd (ClientData info--
*    Accepts a key and a list of data items that will be used to fill a
*    pre-defined structure.
*    The newly filled structure pointer is saved in a hash table,
*    referenced by 'key_Value'.
*
* Arguments:
*   objv[0] "sample"
*   objv[1] "liststr"
*   objv[2] key_Value
*   objv[3] structure_List
* 
* Results:
*   Places a structure pointer in the hash table.
* 
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_ListstrCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    Tcl_HashEntry *hashEntryPtr;
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    Tcl_Obj *returnObjPtr;
    Tcl_Obj *listElementPtr;
    int listlen;
    int  isNew;
    int length;
    char *tmpString;
    int i;
    char *hashEntryContentsPtr;
    char *hashKeyPtr;

    struct demo {
        int	firstInt;
	char    *secondString;
	double  thirdFloat;
    } *demoStruct;
    
    
    debugprt("Sample_ListstrCmd called with %d args\n", objc);

    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = NULL;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /*
     * Allocate space for the structure
     */
    
    demoStruct = (struct demo *) malloc(sizeof (struct demo));
    
    /* 
     * Get the length of the list, then step through it.
     */
    
    Tcl_ListObjLength(interp, objv[3], &listlen);
    
    for(i=0; i<listlen; i++) {
        
	/* 
	 * Extract a list element from the list pointer 
	 */

        Tcl_ListObjIndex(interp, objv[3], i, &listElementPtr);
	debugprt("Position: %d : Value: %s\n", 
	   i, Tcl_GetStringFromObj(listElementPtr, NULL));
        
	/* 
	 * A strange way to determine which structure element is
	 *  being processed, but it works in the loop.
	 * If an illegal value is in the object, the error return
	 *  will be set automaticly by the interpreter.
	 */

	switch (i) {
	case 0: {
	            int t;
		    if (TCL_OK !=
		        Tcl_GetIntFromObj(interp, listElementPtr, &t)) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct->firstInt = t;
		    }
		break;
		}
	case 1: {
	            char *t;
		    if (NULL == (t = 
		        Tcl_GetStringFromObj(listElementPtr, NULL))) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct->secondString= (char *)malloc(strlen(t)+1);
			strcpy(demoStruct->secondString, t);
		    }
		break;
		}
	case 2: {
	            double t;
		    if (TCL_OK !=
		        Tcl_GetDoubleFromObj(interp, listElementPtr, &t)) {
		            returnStructPtr->status = TCL_ERROR;
		            goto done;
		    } else {
		        demoStruct->thirdFloat = t;
		    }
		break;
		}
	}
    }
    
    /* 
     * Extract the string representation of the object argument,
     * and use that as a key into the hash table.
     *
     * If this entry already exists, complain.
     */

    hashKeyPtr = Tcl_GetStringFromObj(objv[2], NULL);
    hashEntryPtr = Tcl_CreateHashEntry(sample_hashtablePtr, hashKeyPtr, &isNew);
    
    if (!isNew) {
        char errString[80];

	sprintf(errString, "Hashed object named \"%s\" already exists.\n", 
	    hashKeyPtr);

        /*
         * Both of these strings will be added to the Tcl script
	 *  global variable errorInfo
	 */

        Tcl_AddObjErrorInfo(interp, errString, strlen(errString));
        Tcl_AddErrorInfo(interp, "error in Sample_CreateCmd");
	
	/*
	 * This SetErrorCode command will set the Tcl script
	 *  variable errorCode to "5 5 5"
	 */

	Tcl_SetErrorCode(interp, "5", "5", "5", (char *) NULL);
	
	/* 
	 * This defines the return string for this subcommand
	 */

        Tcl_AppendResult(interp, "Hashed object named \"",
                hashKeyPtr, "\" already exists.", (char *) NULL);

	returnStructPtr->status = TCL_ERROR;
        goto done;
    }
    
    /*
     * If we are here, then the key is unused.
     * Get the string representation from the object, 
     *  and make a copy that can be placed into the hash table.
     */

    tmpString = Tcl_GetStringFromObj(objv[3], &length);
    hashEntryContentsPtr = (char *) malloc(length+1);
    strcpy(hashEntryContentsPtr, tmpString);
    
    debugprt("setting: %s\n", hashEntryContentsPtr);
    Tcl_SetHashValue(hashEntryPtr, demoStruct);
    
    
done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}



/*------------------------------------------------------------------------
* CmdReturn *Sample_GetstrCmd ()--
*    Demonstrates retrieving a structure pointer from a hash table, and 
*    stuffing that into a list.
*
* Arguments
*   objv[0]:	"sample"
*   objv[1]:	"getstr"
*   objv[2]:	hash_Key
* 
* Results
*   No changes to hash table.  Returns saved value, or sets error.
*
* Side Effects:
*   None
------------------------------------------------------------------------*/
CmdReturn *Sample_GetstrCmd (ClientData info, 
			 Tcl_Interp *interp, 
			 int objc, 
			 Tcl_Obj * CONST *objv) {
    Tcl_HashEntry *hashEntryPtr;
    CmdReturn *returnStructPtr;
    char *returnCharPtr;
    Tcl_Obj *returnObjPtr;
    Tcl_Obj *listPtrPtr[3];
    
    char *hashKeyPtr;

    struct demo {
        int	firstInt;
	char    *secondString;
	double  thirdFloat;
    } *demoStruct;
    
    
    debugprt("Sample_GetCmd called with %d args\n", objc);
    
    /* 
     * Allocate the space and initialize the return structure.
     */

    returnStructPtr = (CmdReturn *) malloc(sizeof (CmdReturn));
    returnStructPtr->status = TCL_OK;
    returnStructPtr->object = NULL;
    returnCharPtr = NULL;
    returnObjPtr    = NULL;
    
    /*
     * Get the key from the argument
     * and attempt to extract that entry from the hashtable.
     * If the returned entry pointer is NULL, this key is not in the table.
     */

    hashKeyPtr = Tcl_GetStringFromObj(objv[2], NULL);

    hashEntryPtr = Tcl_FindHashEntry(sample_hashtablePtr, hashKeyPtr);

    if (hashEntryPtr == (Tcl_HashEntry *) NULL) {
        char errString[80];
	Tcl_Obj *errCodePtr;

	/* 
	 * This string will be returned as the result of the command.
	 */

        Tcl_AppendResult(interp, "can not find hashed object named \"",
                hashKeyPtr, "\"", (char *) NULL);
	returnStructPtr->status = TCL_ERROR;
        goto done;
    }
    
    /* 
     * If we got here, then the search was successful and we can extract
     *  the data value from the hash entry and return it.
     */

    demoStruct = (struct demo *)Tcl_GetHashValue(hashEntryPtr);
    
    /* 
     * Create three objects with the values from the structure,
     * and then merge them into a list object.
     * 
     * Return the list object.
     */
    
    listPtrPtr[0] = Tcl_NewIntObj(demoStruct->firstInt);
    listPtrPtr[1] = Tcl_NewStringObj(demoStruct->secondString, -1);
    listPtrPtr[2] = Tcl_NewDoubleObj((double) demoStruct->thirdFloat);
    
    returnObjPtr = Tcl_NewListObj(3, listPtrPtr);
    
    debugprt("returnString: %s\n", returnCharPtr);

done:
    if ((returnObjPtr == NULL) && (returnCharPtr != NULL)) {
        returnObjPtr = Tcl_NewStringObj(returnCharPtr, -1);
    }
                            
    returnStructPtr->object = returnObjPtr;
    if (returnCharPtr != NULL) {free(returnCharPtr);}

    return returnStructPtr;
}



